import logging
from typing import Any, List

import torch
from torch import nn

from . import layers


class FMGraphNode:
    def __init__(self, value: torch.Tensor, parents=tuple(), depth=0, layer=None):
        """Node of a Functional Model.

        This might represent input or intermediate values of the neural network.

        Parameters
        ----------
        value
            Tensor object representing value of the node.
        parents
            Tuple of parent nodes.
        depth
            How deep the node is in the tree.
        layer
            nn.Module object that transformed parents.
        """
        self._v = value
        self.parents = parents
        self.children: List[FMGraphNode] = []
        self.layer = layer
        self.depth = depth
        self._output = None
        self._parents_outputs: List[Any] = []

    @property
    def channels(self):
        return self._v.shape[1]

    @property
    def features(self):
        return self._v.shape[1]

    @property
    def H(self):
        assert len(self._v.shape) == 4, "Variable is not an image!"
        return self._v.shape[2]

    @property
    def W(self):
        assert len(self._v.shape) == 4, "Variable is not an image!"
        return self._v.shape[3]

    @property
    def shape(self):
        return (None, *self._v.shape[1:])

    @property
    def numel(self):
        return self._v.shape[1:].numel()

    def apply_layer(self, layer, *others):
        if others:
            new_depth = min(self.depth, *(o.depth for o in others)) + 1
        else:
            new_depth = self.depth + 1

        new_output = layer.__call__(self._v, *(o._v for o in others))

        new_layer_node = FMGraphNode(value=new_output, parents=(self, *others), layer=layer, depth=new_depth)
        self.children.append(new_layer_node)
        for other in others:
            other.children.append(new_layer_node)
        return new_layer_node

    def _get_all_nodes_below(self, layer_list):
        if self in layer_list:
            return layer_list
        layer_list.append(self)

        for child in self.children:
            child._get_all_nodes_below(layer_list)
        return layer_list

    def _get_all_nodes_above(self, layer_list):
        if self in layer_list:
            return layer_list
        layer_list.append(self)

        for child in self.parents:
            child._get_all_nodes_above(layer_list)
        return layer_list

    def _remove_outsiders_below(self, insiders, already_called):
        if self in already_called:
            return None
        already_called.append(self)

        for child in self.children.copy():
            if child in insiders:
                child._remove_outsiders_below(insiders, already_called)
            else:
                self.children.remove(child)

    def _forward_edge(self, x):
        if len(self.parents) == 0:
            for child in self.children:
                child._forward_edge(x)
            return

        self._parents_outputs.append(x)

        if len(self._parents_outputs) == len(self.parents):
            self._output = self.layer.__call__(*self._parents_outputs)
            self._parents_outputs = []
            for child in self.children:
                child._forward_edge(self._output)

    def _clear_memory(self):
        self._parents_outputs = []
        self._output = None

    def __call__(self, *args, **kwargs):
        return self.apply_layer(*args, **kwargs)

    def __abs__(self):
        return self.apply_layer(layers.AnyOpLayer(lambda x: abs(x)))

    def __add__(self, other):
        if isinstance(other, FMGraphNode):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self.apply_layer(layers.AddOpLayer(), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x + other))

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, FMGraphNode):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self.apply_layer(layers.MulOpLayer(), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x - other))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.apply_layer(layers.AnyOpLayer(op=lambda x: -x))

    def __pow__(self, other):
        if isinstance(other, FMGraphNode):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self.apply_layer(layers.AnyOpLayer(op=lambda x, y: x**y), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x**other))

    def __rpow__(self, other):
        if isinstance(other, FMGraphNode):
            return other.__pow__(self)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other**x))

    def __sub__(self, other):
        if isinstance(other, FMGraphNode):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self.apply_layer(layers.SubOpLayer(), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x - other))

    def __rsub__(self, other):
        if isinstance(other, FMGraphNode):
            return other.__sub__(self)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other - x))

    def __truediv__(self, other):
        if isinstance(other, FMGraphNode):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self.apply_layer(layers.AnyOpLayer(op=lambda x, y: x / y), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x / other))

    def __rtruediv__(self, other):
        if isinstance(other, FMGraphNode):
            return other.__truediv__(self)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other / x))

    def __matmul__(self, other):
        if isinstance(other, FMGraphNode):
            return self.apply_layer(layers.MatmulOpLayer(), other)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x @ other))

    def __rmatmul__(self, other):
        if isinstance(other, FMGraphNode):
            return other.__matmul__(self)
        else:
            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other @ x))


class Input(FMGraphNode):
    def __init__(
        self,
        shape,
        dtype=torch.float32,
        min_value=0.0,
        max_value=1.0,
        _batch_size=1,
        _use_tensor=None,
    ):
        """Input to the Functional Model.

        It should be treated as a placeholder value that will be replaced with
        real data after the model is created.
        For calculation purposes, it can be treated as a normal numerical object,
        which means it can be added, subtracted, multiplied, taken absolute value of,
        etc.

        Parameters
        ----------
        shape
            Shape of the real data. Should NOT include the batch dimension.
        dtype
            Dtype of the real data that will be the input of the network.
        min_value
            In rare cases, if real world data is very specific and some values
            cannot work with the model, this should be used to set the
            reasonable minimal value that the model will work on.
        max_value
            As above, but the maximal value.
        """
        if _use_tensor is not None:
            super().__init__(value=_use_tensor)
            return

        value = torch.rand(_batch_size, *shape) * (max_value - min_value) + min_value
        value = value.to(dtype)
        super().__init__(value=value)


class FunctionalModel(nn.Module):
    def __init__(self, inputs, outputs):
        """A PyTorch model that applies operations defined by the graph.

        All operations that changed ``inputs`` into ``outputs`` will be applied
        in the same order on the real data that will be fed into this model.

        Parameters
        ----------
        inputs
            FMGraphNode object or a tuple of them.
        outputs
            FMGraphNode object or a tuple of them.
        """
        super().__init__()
        logging.info(f"Creating a Functional Model!")

        if isinstance(inputs, FMGraphNode):
            inputs = (inputs,)
        assert all(isinstance(x, FMGraphNode) for x in inputs)
        self.inputs = inputs

        if isinstance(outputs, FMGraphNode):
            outputs = (outputs,)
        assert all(isinstance(x, FMGraphNode) for x in outputs)
        self.outputs = outputs

        self._has_single_input = len(self.inputs) == 1
        self._has_single_output = len(self.outputs) == 1
        self._registered_modules = []
        self._prune_unused_layers()
        self._register_reachable_modules()

    def forward(self, inputs):
        if self._has_single_input:
            self.inputs[0]._forward_edge(inputs)
        else:
            assert len(inputs) == len(self.inputs), "Number of inputs doesn't match!"
            for root, arg in zip(self.inputs, inputs):
                root._forward_edge(arg)

        if self._has_single_output:
            return self.outputs[0]._output
        else:
            return tuple(output_leaf._output for output_leaf in self.outputs)

    @property
    def input_shape(self):
        if self._has_single_input:
            return self.inputs[0].shape
        else:
            return tuple(node.shape for node in self.inputs)

    @property
    def output_shape(self):
        if self._has_single_output:
            return self.outputs[0].shape
        else:
            return tuple(node.shape for node in self.outputs)

    def _register_module(self, node):
        if not isinstance(node.layer, nn.Module) or node.layer in self._registered_modules:
            return False

        num_modules = len(self._registered_modules)
        self.add_module(name=f"module{num_modules:0>3}_depth{node.depth:0>3}", module=node.layer)
        self._registered_modules.append(node.layer)
        return True

    def _register_reachable_modules(self):
        reachable_nodes = self._reachable_nodes
        num_registered = 0
        for node in reachable_nodes:
            if self._register_module(node):
                num_registered += 1
        logging.info(f"Registered {num_registered} modules!")

    def _prune_unused_layers(self):
        used_nodes = self._used_nodes
        used_nodes_ids = {id(node) for node in used_nodes}
        reachable_nodes_ids = {id(node) for node in self._reachable_nodes}
        to_prune = reachable_nodes_ids.difference(used_nodes_ids)

        logging.info(f"Pruning {len(to_prune)} modules!")

        already_called = []
        for root in self.inputs:
            root._remove_outsiders_below(insiders=used_nodes, already_called=already_called)

    def _clear_nodes_memory(self):
        for node in self._reachable_nodes:
            node._clear_memory()

    @property
    def _reachable_nodes(self):
        nodes_below = []
        for root in self.inputs:
            root._get_all_nodes_below(nodes_below)
        return nodes_below

    @property
    def _used_nodes(self):
        nodes_above = []
        for output_leaf in self.outputs:
            output_leaf._get_all_nodes_above(nodes_above)
        return nodes_above



def get_parameter_count(model):
    cnt = 0
    for param in model.parameters():
        cnt += param.shape.numel()
    return cnt


def get_parameter_shapes(model):
    shapes = []
    for param in model.parameters():
        shapes.append(tuple(param.shape))
    return shapes


def model_similar(a, b):
    if get_parameter_count(a) != get_parameter_count(b):
        return False

    if sorted(get_parameter_shapes(a)) != sorted(get_parameter_shapes(b)):
        return False
    return True



import torch

from .. import examples, tools


def test_toy_resnet():
    model = examples.ToyResNet((3, 32, 32), n_classes=10)
    inputs = torch.rand(16, 3, 32, 32)
    outputs = model(inputs)
    assert list(outputs.shape) == [16, 10]
    assert tools.get_parameter_count(model) == 175178


def test_wrn():
    model = examples.ResNet((3, 32, 32), n_classes=10, version=("WRN", 16, 4))
    inputs = torch.rand(16, 3, 32, 32)
    outputs = model(inputs)
    assert list(outputs.shape) == [16, 10]
    assert tools.get_parameter_count(model) == 2750698


def test_vgg():
    model = examples.VGG((3, 32, 32), n_classes=10, version=13)
    inputs = torch.rand(16, 3, 32, 32)
    outputs = model(inputs)
    assert list(outputs.shape) == [16, 10]
    assert tools.get_parameter_count(model) == 9413066


def test_enc_dec():
    model = examples.simple_encoder_decoder((3, 32, 32))
    inputs = torch.rand(16, 3, 32, 32)
    outputs = model(inputs)
    assert list(outputs.shape) == [16, 1, 7, 7]
    assert tools.get_parameter_count(model) == 28529



"""
This is a flexible implementation of ResNet architecture.
It allows for creation of standard ResNet v2 or Wide ResNet variants.
"""

from torch import nn

from .. import FunctionalModel, Input
from .common import classifier


def shortcut_func(x, channels, stride):
    if x.channels != channels or stride != 1:
        return x(nn.Conv2d(x.channels, channels, kernel_size=1, bias=False, stride=stride))
    else:
        return x


def ToyResNet(input_shape, n_classes):
    """A basic example of ResNet with low degree of customizability."""

    inputs = Input(input_shape)
    flow = inputs(nn.Conv2d(inputs.channels, 16, 3, 1, 1))

    for group_size, width, stride in zip((2, 2, 2), (16, 32, 64), (1, 2, 2)):
        for _ in range(group_size):
            shortcut = shortcut_func(flow, width, stride)
            flow(nn.BatchNorm2d(flow.features))(nn.ReLU())
            flow = flow(nn.Conv2d(flow.channels, width, 3, stride, 1))
            flow = flow(nn.BatchNorm2d(flow.features))(nn.ReLU())
            flow = flow(nn.Conv2d(flow.channels, width, 3, 1, 1))

            flow = flow + shortcut
            stride = 1
    flow = flow(nn.BatchNorm2d(flow.features))(nn.ReLU())
    outs = classifier(flow, n_classes, pooling="avgpool")
    model = FunctionalModel(inputs=inputs, outputs=outs)
    return model


def ResNet(
    input_shape,
    n_classes,
    version=None,
    bootleneck=False,
    strides=(1, 2, 2),
    group_sizes=(2, 2, 2),
    channels=(16, 32, 64),
    activation=nn.ReLU(),
    final_pooling="avgpool",
    dropout=0,
    bn_ends_block=False,
    **kwargs,
):
    if version:
        if version == 20:
            group_sizes = (3, 3, 3)
        elif version == 32:
            group_sizes = (5, 5, 5)
        elif version == 44:
            group_sizes = (7, 7, 7)
        elif version == 56:
            group_sizes = (9, 9, 9)
        elif version == 110:
            group_sizes = (18, 18, 18)
        elif version == 164:
            bootleneck = True
            channels = (64, 128, 256)
            group_sizes = (18, 18, 18)
        elif isinstance(version, tuple) and version[0] == "WRN":
            _, N, K = version
            assert (N - 4) % 6 == 0, "N-4 has to be divisible by 6"
            lpb = (N - 4) // 6  # layers per block
            group_sizes = (lpb, lpb, lpb)
            channels = tuple(c * K for c in channels)
        else:
            raise NotImplementedError(f"Unkown version={version}!")
    if kwargs:
        print(f"ResNet: unknown parameters: {kwargs.keys()}")

    def simple_block(flow, channels, stride):
        if preactivate_block:
            flow = flow(nn.BatchNorm2d(flow.features))(activation)

        flow = flow(nn.Conv2d(flow.channels, channels, 3, stride, 1))
        flow = flow(nn.BatchNorm2d(flow.features))(activation)

        if dropout:
            flow = flow(nn.Dropout(p=dropout))
        flow = flow(nn.Conv2d(flow.channels, channels, 3, 1, 1))

        if bn_ends_block:
            flow = flow(nn.BatchNorm2d(flow.features))(activation)
        return flow

    def bootleneck_block(flow, channels, stride):
        if preactivate_block:
            flow = flow(nn.BatchNorm2d(flow.features))(activation)

        flow = flow(nn.Conv2d(flow.channels, channels // 4, 1))
        flow = flow(nn.BatchNorm2d(flow.features))(activation)

        flow = flow(nn.Conv2d(flow.channels, channels // 4, 3, stride=stride, padding=1))
        flow = flow(nn.BatchNorm2d(flow.features))(activation)

        flow = flow(nn.Conv2d(flow.channels, channels, 1))
        if bn_ends_block:
            flow = flow(nn.BatchNorm2d(flow.features))(activation)
        return flow

    if bootleneck:
        block = bootleneck_block
    else:
        block = simple_block

    inputs = Input(input_shape)

    # BUILDING HEAD OF THE NETWORK
    flow = inputs(nn.Conv2d(inputs.channels, 16, 3, 1, 1))

    # BUILD THE RESIDUAL BLOCKS
    for group_size, width, stride in zip(group_sizes, channels, strides):
        flow = flow(nn.BatchNorm2d(flow.features))(activation)
        preactivate_block = False

        for _ in range(group_size):
            residual = block(flow, width, stride)
            shortcut = shortcut_func(flow, width, stride)
            flow = residual + shortcut
            preactivate_block = True
            stride = 1

    # BUILDING THE CLASSIFIER
    flow = flow(nn.BatchNorm2d(flow.features))(activation)
    outs = classifier(flow, n_classes, pooling=final_pooling)
    model = FunctionalModel(inputs=inputs, outputs=outs)
    return model



from torch import nn

from .. import FunctionalModel, Input, layers, tools


def test_1():
    inputs = Input(shape=(3, 32, 32))
    x = inputs(nn.Conv2d(inputs.channels, 32, 3))(nn.ReLU())
    x = x(nn.Conv2d(x.channels, 64, 3))(nn.ReLU())
    block_1_output = x(nn.MaxPool2d(3))

    x = block_1_output(nn.Conv2d(block_1_output.channels, 64, 3, padding=1))(nn.ReLU())
    x = x(nn.Conv2d(x.channels, 64, 3, padding=1))(nn.ReLU())
    block_2_output = x + block_1_output

    x = block_2_output(nn.Conv2d(block_2_output.channels, 64, 3, padding=1))(nn.ReLU())
    x = x(nn.Conv2d(x.channels, 64, 3, padding=1))(nn.ReLU())
    block_3_output = x + block_2_output

    x = block_3_output(nn.Conv2d(x.channels, 64, 3))(nn.ReLU())
    x = x(nn.AvgPool2d(kernel_size=(x.H, x.W)))(nn.Flatten())
    x = x(nn.Linear(x.features, 256))(nn.ReLU())
    x = x(nn.Dropout(0.5))
    outputs = x(nn.Linear(x.features, 10))

    model = FunctionalModel(inputs, outputs)
    assert tools.get_parameter_count(model) == 223242


def test_2():
    inputs = Input((3, 128, 128))
    x = inputs

    x = x(nn.Conv2d(in_channels=x.channels, out_channels=16, kernel_size=3))
    x = x(nn.MaxPool2d(kernel_size=2))
    x = x(nn.ReLU())

    x = x(nn.Conv2d(in_channels=x.channels, out_channels=32, kernel_size=3))
    x = x(nn.MaxPool2d(kernel_size=2))
    x = x(nn.ReLU())

    x = x(nn.Conv2d(in_channels=x.channels, out_channels=64, kernel_size=3))
    x = x(nn.MaxPool2d(kernel_size=2))
    x = x(nn.ReLU())

    x = x(nn.Conv2d(in_channels=x.channels, out_channels=64, kernel_size=3))
    x = x(nn.MaxPool2d(kernel_size=2))
    x = x(nn.ReLU())

    x = x(nn.Flatten())
    outputs = x(nn.Linear(in_features=x.features, out_features=10))
    model = FunctionalModel(inputs=inputs, outputs=outputs)
    assert model.output_shape == (None, 10)
    assert tools.get_parameter_count(model) == 83562


def test_3():
    task1_input = Input(shape=(1, 28, 28))
    task2_input = Input(shape=(3, 32, 32))

    x = task1_input
    x = x(nn.Conv2d(x.channels, 16, 3))
    x = x(nn.MaxPool2d(3))(nn.ReLU())
    x = x(nn.Flatten())
    head1_out = x(nn.Linear(x.features, 200))

    x = task2_input
    x = x(nn.Conv2d(x.channels, 16, 3))
    x = x(nn.MaxPool2d(3))(nn.ReLU())
    x = x(nn.Flatten())
    head2_out = x(nn.Linear(x.features, 200))

    x = head1_out + head2_out
    x = x(nn.Linear(x.features, 400))(nn.ReLU())
    task1_outputs = x(nn.Linear(x.features, 10))
    task2_outputs = x(nn.Linear(x.features, 10))

    model = FunctionalModel(inputs=(task1_input, task2_input), outputs=(task1_outputs, task2_outputs))
    assert tools.get_parameter_count(model) == 614228


def test_4():
    x1 = Input(shape=(1, 2, 3))
    x2 = Input(shape=(5, 2, 3))
    x = x1(layers.ConcatLayer(dim=1), x2)
    assert x.shape == (None, 6, 2, 3)



"""
This is a flexible implementation of VGG architecture.
"""

from torch import nn

from .. import FunctionalModel, Input
from .common import classifier


def VGG(
    input_shape,
    n_classes,
    version=None,
    group_sizes=(1, 1, 2, 2, 2),
    channels=(64, 128, 256, 512, 512),
    pools=(2, 2, 2, 2, 2),
    activation=nn.ReLU(),
    final_pooling="avgpool",
    **kwargs,
):
    if kwargs:
        print(f"VGG: unknown parameters: {kwargs.keys()}")
    if version:
        if version == 11:
            group_sizes = (1, 1, 2, 2, 2)
        elif version == 13:
            group_sizes = (2, 2, 2, 2, 2)
        elif version == 16:
            group_sizes = (2, 2, 3, 3, 3)
        elif version == 19:
            group_sizes = (2, 2, 4, 4, 4)
        else:
            raise NotImplementedError(f"Unkown version={version}!")

    inputs = Input(shape=input_shape)
    flow = inputs

    iteration = 0
    for group_size, width, pool in zip(group_sizes, channels, pools):
        if iteration == 0:
            iteration = 1
        else:
            flow = flow(nn.MaxPool2d(pool))

        for _ in range(group_size):
            flow = flow(nn.Conv2d(flow.channels, width, 3, 1, 1, bias=False))
            flow = flow(nn.BatchNorm2d(flow.features))(activation)

    outs = classifier(flow, n_classes, final_pooling)
    model = FunctionalModel(inputs=inputs, outputs=outs)
    return model



"""
CORRESPONDES TO THE EXAMPLE FROM TENSORFLOW GUIDE
https://www.tensorflow.org/guide/keras/functional
"""

from torch import nn

from .. import FunctionalModel, Input, layers


def simple_encoder_decoder(input_shape=(1, 28, 28)):
    relu = nn.ReLU()

    encoder_input = x = Input(shape=input_shape)
    x = x(nn.Conv2d(x.channels, 16, 3))(relu)
    x = x(nn.Conv2d(x.channels, 32, 3))(relu)
    x = x(nn.MaxPool2d(3))
    x = x(nn.Conv2d(x.channels, 32, 3))(relu)
    x = x(nn.Conv2d(x.channels, 16, 3))(relu)
    encoder_output = x(nn.MaxPool2d(kernel_size=(x.H, x.W)))(nn.Flatten())
    encoder = FunctionalModel(inputs=encoder_input, outputs=encoder_output)

    decoder_input = x = Input(shape=(encoder_output.features,))
    x = x(layers.ReshapeLayer((1, 4, 4)))
    x = x(nn.ConvTranspose2d(x.channels, 16, 3))(relu)
    x = x(nn.ConvTranspose2d(x.channels, 32, 3))(relu)
    x = x(nn.Upsample(3))
    x = x(nn.ConvTranspose2d(x.channels, 16, 3))(relu)
    decoder_output = x(nn.ConvTranspose2d(x.channels, 1, 3))(relu)
    decoder = FunctionalModel(inputs=decoder_input, outputs=decoder_output)

    autoencoder_input = Input(shape=input_shape)
    encoded_img = autoencoder_input(encoder)
    decoded_img = encoded_img(decoder)
    autoencoder = FunctionalModel(autoencoder_input, decoded_img)
    return autoencoder


if __name__ == "__main__":
    from logging import DEBUG, basicConfig

    import torch

    from pytorch_functional import tools

    basicConfig(level=DEBUG)

    model = simple_encoder_decoder((1, 28, 28))

    inputs = torch.rand(1, 1, 28, 28)
    outs = model.forward(inputs)
    print(f"Parameters: {tools.get_parameter_count(model)}")



import torch
from torch import nn


class AddOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a + b


class SubOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a - b


class MulOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a * b


class MatmulOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a @ b


class AnyOpLayer(nn.Module):
    def __init__(self, op):
        super().__init__()
        self.forward = op


class ConcatLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, *elements):
        return torch.cat(tensors=elements, dim=self.dim)


class StackLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, *elements):
        return torch.stack(tensors=elements, dim=self.dim)


class ReshapeLayer(nn.Module):
    def __init__(self, shape, batch_size_included=False):
        super().__init__()
        if not batch_size_included:
            self.shape = (-1, *shape)
        else:
            self.shape = shape

    def forward(self, inputs):
        return torch.reshape(input=inputs, shape=self.shape)



import inspect
import logging

from torch import nn

from .functional_model import FMGraphNode

_objects_with_wrapped_call = []


def enable_functional_api_for_module(module):
    """Enable functional API.

    This means:
        * create __pytorch_functional_old_call__ that holds old call
        * wrap __call__
    """
    assert inspect.isclass(module), "Argument must be a class!"
    logging.debug(f"ENABLING FUNCTIONAL API FOR {module}")

    assert "__pytorch_functional_old_call__" not in vars(
        module
    ), f"Functional API already enabled for {module}!"

    __old_call__ = module.__call__
    module.__pytorch_functional_old_call__ = __old_call__

    def experimental_monkey_patch_call(self, *args, **kwds):
        if len(args) > 0 and len(kwds) == 0 and all((isinstance(x, FMGraphNode) for x in args)):
            node = args[0]
            return node.apply_layer(self, *args[1:])
        else:
            return __old_call__(self, *args, **kwds)

    _objects_with_wrapped_call.append(module)
    module.__call__ = experimental_monkey_patch_call


def disable_functional_api_for_module(module):
    logging.debug(f"DISABLING EXPERIMENTAL API FOR {module}")

    assert hasattr(module, "__pytorch_functional_old_call__"), f"Functional API not enabled for {module}!"
    module.__call__ = module.__pytorch_functional_old_call__
    del module.__pytorch_functional_old_call__


def functional_api_new_wrapper(self, *args, **kwds):
    obj = super(nn.Module, self).__new__(self)
    if type(obj) not in _objects_with_wrapped_call:
        enable_functional_api_for_module(type(obj))
    return obj


def functional_api_new_wrapper_backup(self, *args, **kwds):
    obj = super(nn.Module, self).__new__(self)
    return obj


def enable_functional_api_for_new_modules():
    nn.Module.__new__ = functional_api_new_wrapper


def disable_functional_api_for_new_modules():
    nn.Module.__new__ = functional_api_new_wrapper_backup


def remove_call_wrapper_from_all_modules():
    global _objects_with_wrapped_call
    for obj in _objects_with_wrapped_call:
        disable_functional_api_for_module(obj)
    _objects_with_wrapped_call = []



from torch import nn

from .. import layers


def classifier(flow, n_classes, pooling="avgpool"):
    if pooling == "catpool":
        maxp = flow(nn.MaxPool2d(kernel_size=(flow.H, flow.W)))
        avgp = flow(nn.AvgPool2d(kernel_size=(flow.H, flow.W)))
        flow = maxp(layers.ConcatLayer(dim=1), avgp)(nn.Flatten())
    if pooling == "avgpool":
        flow = flow(nn.AvgPool2d(kernel_size=(flow.H, flow.W)))(nn.Flatten())
    if pooling == "maxpool":
        flow = flow(nn.MaxPool2d(kernel_size=(flow.H, flow.W)))(nn.Flatten())
    return flow(nn.Linear(flow.features, n_classes))



import inspect
import logging

from torch import nn

from . import layers
from .functional_model import FMGraphNode

logging.debug("STARTING DETECTING MODULES")

predefined_modules = []
for value in nn.__dict__.values():
    if inspect.isclass(value) and issubclass(value, nn.Module):
        logging.debug(f"DETECTED {value}")
        predefined_modules.append(value)
for value in layers.__dict__.values():
    if inspect.isclass(value) and issubclass(value, nn.Module):
        logging.debug(f"DETECTED {value}")
        predefined_modules.append(value)

logging.debug("DETECTING MODULES FINISHED")


def enable_functional_api_for_package(pkg):
    """Search through the package and enable functional API for every nn.Module."""
    for value in pkg.__dict__.values():
        if inspect.isclass(value) and issubclass(value, nn.Module):
            logging.debug(f"DETECTED {value}")
            enable_functional_api_for_module(value)


def disable_functional_api_for_package(pkg):
    """Search through the package and disable functional API for every nn.Module."""
    for value in pkg.__dict__.values():
        if inspect.isclass(value) and issubclass(value, nn.Module):
            logging.debug(f"DETECTED {value}")
            disable_functional_api_for_module(value)


def enable_functional_api_for_module(module):
    """Enable functional API.

    This means:
        * create __pytorch_functional_old_call__ that holds old call
        * wrap __call__
    """
    logging.debug(f"ENABLING EXPERIMENTAL API FOR {module}")

    assert "__pytorch_functional_old_call__" not in vars(
        module
    ), f"Functional API already enabled for {module}!"

    __old_call__ = module.__call__
    module.__pytorch_functional_old_call__ = __old_call__

    def experimental_monkey_patch_call(self, *args, **kwds):
        if len(args) > 0 and len(kwds) == 0 and all((isinstance(x, FMGraphNode) for x in args)):
            node = args[0]
            return node.apply_layer(self, *args[1:])
        else:
            return __old_call__(self, *args, **kwds)

    module.__call__ = experimental_monkey_patch_call


def disable_functional_api_for_module(module):
    logging.debug(f"DISABLING EXPERIMENTAL API FOR {module}")

    assert hasattr(module, "__pytorch_functional_old_call__"), f"Functional API not enabled for {module}!"
    module.__call__ = module.__pytorch_functional_old_call__
    del module.__pytorch_functional_old_call__


def enable_experimental_api_for_predefined_modules():
    for module in predefined_modules:
        enable_functional_api_for_module(module)


def disable_experimental_api_for_predefined_modules():
    for module in predefined_modules:
        disable_functional_api_for_module(module)


def enable_experimental_api():
    enable_experimental_api_for_predefined_modules()
